package xyz.matj1.bp;

import mcdr.objects.utils.RankedSortedCollectionDistHashcode;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import messif.operations.AnswerType;
import messif.operations.query.KNNQueryOperation;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtils;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.Range;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import xyz.matj1.bp.Utils.Pair;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BiFunction;

;


/**
 * A class for the testing of how sequences of
 * motion words of similar motions of different lengths can be compared
 *
 * @author Matěj Bagar
 */
public class LengthTesting {
	/**
	 * This searches for the closes sequences of each sequence
	 * and counts how many of them are from the same category.
	 *
	 * Tato třída je pro testování motion words tak,
	 * že se vyhledávají nejpodobnější pohyby k danému pohybu s určitou délkou,
	 * a měří se, jak pravděpodobně jsou mezi těmi nejpodobnějšími pohyby
	 * různě dlouhé pohyby ze stejné kategorie. (starý popis této třídy)
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @return a Map of a Category and a List of Pairs of each object in the category and its score
	 */
	public static Map<Category, List<Pair<LocalAbstractObject, Double>>> knnLengthTesting(ObjectMgmt mwMgmt) {
		var categoryObjects = mwMgmt.getCategoryObjects();
		var accumulator = new ConcurrentHashMap<
			Category,
			List<Pair<LocalAbstractObject, Double>>
		>();

		var  dataStream = categoryObjects.entrySet().parallelStream();
		dataStream.forEach(entry -> {
			var category = entry.getKey();
			var sequences = entry.getValue();
			accumulator.put(category, new ArrayList<>());

			for (var sequence : sequences) {
				var operation = new KNNQueryOperation(
					sequence,
					sequences.size() - 1,
					false,
					AnswerType.ORIGINAL_OBJECTS,
					new RankedSortedCollectionDistHashcode()
				);

				mwMgmt.getObjects().forEach(object -> {
					if (!object.getLocatorURI().equals(sequence.getLocatorURI())) {
						operation.addToAnswer(object);
					}
				});

				var answer =
					Utils.iteratorToStream(operation.getAnswer())
						.map(o -> (LocalAbstractObject)(o.getObject())); // Proč KNNOperation bere LocalAbstractObject a vrací DistanceRankedObject<>?

				var answerScore = answer.mapToInt(o ->
					mwMgmt
						.getObjectCategories(o)
						.contains(category)
						? 1 : 0
				).sum() / (double)operation.getK();

				accumulator.get(category).add(new Pair<>(sequence, answerScore));
		}});

		return accumulator;
	}

	/**
	 * Creates an XY graph of the length and the KNN score for each sequence in the category for each category.
	 * That means that the points of each category are shown in their own graph, separated from the points of other categories.
	 *
	 * This uses caching to a file currently hardcoded to „cache/knnLengthTesting“ because the computation is long.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param graphsPath path of the directory where the graphs shall be saved
	 * @throws IOException if there is an error in writing to a file in graphsPath
	 */
	public static void knnLengthTestingShow(ObjectMgmt mwMgmt, File graphsPath) throws IOException {
		var results = knnLengthTesting(mwMgmt);

		for (var entry : results.entrySet()) {
			var category = entry.getKey();
			var values = entry.getValue();

			var dataset = new XYSeriesCollection();
			var dataSeries = new XYSeries(category.index);

			values.forEach(p -> dataSeries.add(
				Utils.getLength(p.first()),
				p.second()
			));

			dataset.addSeries(dataSeries);

			var chart = ChartFactory.createScatterPlot(
				"The length and the KNN search score for each sequence in category " + category.id,
				"the length of the sequence",
				"the kNN precision",
				dataset,
				PlotOrientation.VERTICAL,
				false,
				false,
				false
			);
			chart.getXYPlot().getRangeAxis().setRange(0, 1);

			var graphFile = new File(graphsPath, "graph" + category.id + ".png");
			ChartUtils.saveChartAsPNG(graphFile, chart, 1024, 1024);
		}

		System.err.println("KNN score / length graphs for " + mwMgmt + " written to " + graphsPath.getPath());
	}

	/**
	 * Calculates the length relation and the distance for each pair of sequences in each category.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 * @return a map of categories to maps of pairs of the sequences to the results
	 */
	public static Map<Category, Map<Pair<LocalAbstractObject, LocalAbstractObject>, Pair<Double, Double>>> pairwiseCorrelation(
		ObjectMgmt mwMgmt,
		Utils.LengthRelation lengthRelation
	) {
		var categoryObjects = mwMgmt.getCategoryObjects();
		var accumulator = new HashMap<
			Category,
			Map<
				Pair<LocalAbstractObject, LocalAbstractObject>,
				Pair<Double, Double>
		>>();

		for (var entry : categoryObjects.entrySet()) {
			var category = entry.getKey();
			var sequences = entry.getValue();
			accumulator.put(category, new HashMap<>());

			for (int i1 = 0; i1 < sequences.size(); i1++) {
				for (int i2 = i1+1; i2 < sequences.size(); i2++) {
					var seq1 = sequences.get(i1);
					var seq2 = sequences.get(i2);

					if (category.id.equals("73") && Utils.normaliseQuotient((double)Utils.getLength(seq1) / Utils.getLength(seq2)) > 10) {
						continue;
					} // to skip anomalous data points caused by an error in the data set

					BiFunction<Double, Double, Double> relationOperation = switch (lengthRelation) {
						case DIFFERENCE -> (a, b) -> Math.abs(a - b);
						case QUOTIENT -> (a, b) -> Utils.normaliseQuotient(a / b);
					};
					accumulator.get(category).put(
						new Pair<>(seq1, seq2),
						new Pair<>(
							relationOperation.apply((double)Utils.getLength(seq1), (double)Utils.getLength(seq2)),
							(double)seq1.getDistance(seq2)
					));
		}}}

		System.err.println("Pairwise correlations for " + mwMgmt + " calculated");
		return accumulator;
	}

	/**
	 * Creates an XY graph of the length relation and the distance for each pair of sequences in the category for each category.
	 * That means that the points of each category are shown in their own graph, separated from the points of other categories.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 * @param graphsPath path of the directory where the graphs shall be saved
	 * @throws IOException if there is an error in writing to a file in graphsPath
	 */
	public static void pairwiseCorrelationShow(
		ObjectMgmt mwMgmt,
		Utils.LengthRelation lengthRelation,
		File graphsPath
	) throws IOException {
		if (!graphsPath.isDirectory()) {
			throw new IllegalArgumentException(graphsPath.getPath() + " isn't a directory.");
		}

		var pairwiseCorrelationMap = pairwiseCorrelation(mwMgmt, lengthRelation);

		for (var entry : pairwiseCorrelationMap.entrySet()) {
			var dataset = new XYSeriesCollection();

			var category = entry.getKey();
			var values = entry.getValue();
			var dataSeries = new XYSeries(category.index);
			values.values().forEach(v -> dataSeries.add(
				(double)switch (lengthRelation) {
					case QUOTIENT -> v.first(); //Math.log(v.first()) / Math.log(2);
					default -> v.first();
				},
				v.second()
			));
			dataset.addSeries(dataSeries);

			var chart = ChartFactory.createScatterPlot(
				"The length relation and the distance of each two sequences in category " + category.id,
				"the " + switch(lengthRelation) {
					case DIFFERENCE -> "difference";
					case QUOTIENT -> "quotient";
				} +" of the lengths",
				"the distance",
				dataset,
				PlotOrientation.VERTICAL,
				false,
				false,
				false
			);

			var graphFile = new File(graphsPath, "graph" + category.id + ".png");
			ChartUtils.saveChartAsPNG(graphFile, chart, 1024, 1024);
		}

		System.err.println("Pairwise correlations for " + mwMgmt + " written to " + graphsPath.getPath());
	}
}
