package xyz.matj1.bp;

import mcdr.objects.impl.ObjectMocapPoseCoordsL2;
import mcdr.sequence.impl.SequenceMocapPoseCoordsL2DTW;
import mcdr.test.utils.ObjectCategoryMgmt;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;


/**
 * Třída pro měnění rychlosti motion-capture dat pro účely testování motion words
 *
 * @author Matěj Bagar
 */
public class LinearTimeWarp {
	public static void main(String[] args) throws IOException {
		// load data
		var mocapMgmt = BPBagar2.loadData(
			"data/class130-actions-coords_normPOS.data",
			SequenceMocapPoseCoordsL2DTW.class
		);
		var categoryObjects = convertMocap(mocapMgmt.getCategoryObjects());

		// stretching of all sequences
		Instant stretchAllTimer = Instant.now();
		var categorySpedUpObjects = stretchCategoryObjects(categoryObjects, 2.0);
		System.out.println("time to stretch all sequences: " + Duration.between(stretchAllTimer, Instant.now()).toMillis() + " ms");

		// save to file
		try(var output = new FileOutputStream("data/class130-actions-coords_normPOS_stretched-2×.data")) {
			for (var entry : categorySpedUpObjects.entrySet()) {
				for (var sequence : entry.getValue()) {
					sequence.write(output);
		}}}
	}

	/**
	 * Převede sequence typu LocalAbstractObject v dané struktuře do sequencí typu SequenceMocapPoseCoordsL2DTW.
	 * V podstatě je to casting, ale uzavřený v malé funkci, aby se snadno hledaly chyby.
	 *
	 * @param original původní struktura s pósami typu LocalAbstractObject
	 * @return nová struktura s pósami typu SequenceMocapPoseCoordsL2DTW
	 */
	private static Map<Category, List<SequenceMocapPoseCoordsL2DTW>> convertMocap(Map<Category, List<LocalAbstractObject>> original) {
		var convertedMap = new HashMap<Category, List<SequenceMocapPoseCoordsL2DTW>>();
		for (var entry: original.entrySet()) {
			var convertedList = new ArrayList<SequenceMocapPoseCoordsL2DTW>();
			for (var object : entry.getValue()) {
				convertedList.add((SequenceMocapPoseCoordsL2DTW) object);
			}
			convertedMap.put(entry.getKey(), convertedList);
		}
		return convertedMap;
	}

	/**
	 * Tato methoda vytvoří sequenci pós, multiple× pomalejší verzi té zadané tím,
	 * že některé pósy vynechá nebo zkopíruje vícekrát.
	 * Pokud multiple je menší než 1, výsledek je kratší (rychlejší) než původní sequence.
	 * Pokud multiple je větší než 1, výsledek je delší (pomalejší) než původní sequence.
	 * Ale pokud multiple je NaN nebo není větší než 0, tato methoda vyhodí IllegalArgumentException.
	 *
	 * @param sequence původní sequence
	 * @param multiple kolikkrát pomalejší než původní sequence má být výsledná sequence
	 * @return pomalejší verze původní sequence
	 * @throws IllegalArgumentException pokud multiple je NaN nebo není větší než 0
	 */
	public static SequenceMocapPoseCoordsL2DTW stretchSequence(
		SequenceMocapPoseCoordsL2DTW sequence,
		Double multiple
	) throws IllegalArgumentException {
		if (!Double.isFinite(multiple) || multiple > 0) {
			throw new IllegalArgumentException("The multiple is not both finite and greater than 0.");
		}

		var stretchedSequenceList = new ArrayList<ObjectMocapPoseCoordsL2>();
		var counter = 0.0;

		for (var item : sequence.getObjects()) {
			while (counter < multiple) {
				stretchedSequenceList.add(item);
				counter++;
			}
			counter -= multiple;
		}

		return new SequenceMocapPoseCoordsL2DTW(stretchedSequenceList);
	}

	/**
	 * Tato methoda vytvoří mapu kategorií a seznamů sequencí pós,
	 * kde každá pósa je multiple× delší než ta odpovídající v zadané mapě.
	 * Pokud multiple je menší než 1, výsledné sequence jsou kratší (rychlejší) než původní sequence.
	 * Pokud multiple je větší než 1, výsledné sequence jsou delší (pomalejší) než původní sequence.
	 * Ale pokud multiple je NaN nebo není větší než 0, tato methoda vyhodí IllegalArgumentException.
	 *
	 * @param categoryObjects původní mapa
	 * @param multiple kolikkrát delší než původní sequence má být každá sequence ve výsledné mapě
	 * @return mapa s nataženými sequencemi
	 * @throws IllegalArgumentException pokud multiple je NaN nebo není větší než 0
	 */
	public static Map<Category, List<SequenceMocapPoseCoordsL2DTW>> stretchCategoryObjects(
		Map<Category, List<SequenceMocapPoseCoordsL2DTW>> categoryObjects,
		Double multiple
	) throws IllegalArgumentException {
		return categoryObjects.entrySet().parallelStream().map(
			e -> new SimpleImmutableEntry<>(
				e.getKey(),
				e.getValue().parallelStream().map(
					l -> stretchSequence(l, multiple)
				).collect(Collectors.toList())
		)).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
	}
}
